function [flag, metric] = ferrariCanny(Gn, B, plt)
% This function calculates Ferrari-Canny form closure metric for grasps of
% objects with any number of degrees of freedom.  The metric is the
% radius of the largest inscribed sphere in the polyhedron formed by the
% convex sum of the columns of the grasp matrix.
%
% Two cases are handled: (1) the object is constrained only by unilateral
% contacts (captured by Gn) and (2) the object is constrained also by 
% bilateral constraints (captured by B).  The analysis is done on the
% matrix G, which is computed from Gn and B and plays the role of Gn after
% the bilateral constraints are accounted for.
%
%   usage:  [flag, metric] = ferrariCanny(Gn, B, plt)
%
% Input parameters:
%     Gn  - matrix whose columns are the unit normal wrenches. 
%                 Unilateral constraints:   Gn'*nu >= 0
%     B   - matrix whose columns are the unit wrenches for bilateral
%                 constraints.  Bilateral constraints:     B'*nu = 0
%     plt - boolean to switch plotting on or off:
%         = false --> no plotting
%         = true  --> plot (2d and 3d plots only)
%
% Output parameters:
%     flag = -2  -->  grasp doesn't have enough contacts for form closure
%          = -1  -->  Gn not full row rank, so form closure doesn't exist
%          =  0  -->  grasp does not have form closure 
%          =  1  -->  grasp has form closure
%     metric - the radius of the largest inscribed sphere whose center is
%                   at the origin of the wrench space

% Written by Li (Emma) Zhang and Jeff Trinkle, Feburary, 2013

% Check for wrong number of inputs.  We handle too few.  Matlab handles too
% many.
if nargin == 0
    error('ferrariCanny requires 1, 2, or 3 arguments.  You passed 0.');
end

% Get the number of rows and columns of Gn
% nnu --> number of degrees of freedom of object
% ncU --> number of unilateral contacts
[nnu, ncU] = size(Gn);
if nnu == 0
    error('You passed an empty wrench matrix, Gn, into formClosure()');
end
flag = 0;   % Assume no form closure
metric = 0; % Assume no form closure

if nargin == 1
    % B is empty and no plotting is requested
    plt = false;
    G = Gn;       % G is Gn.  No additional constraints imposed by B
    ndof_lost = 0;
end
if nargin == 2
    plt = false;
end
if nargin == 2  ||  nargin == 3
    % B has been passed in
    [nrB, ncB] = size(B);
    if nrB == 0
        % No bilateral constraints imposed.
        G = Gn;
        ndof_lost = 0;
    else
        if nrB ~= nnu
            error('Input matrices must have the same number of rows!');
        else
            ndof_lost = rank(B);
            % A_B is defined as the transpose of the matrix whose
            % columns span the null space of B' and thus A_B*B = 0.
            A_B = null(B')';
            % This is the grasp matrix \overline{G} defined in the Handbook
            G = A_B*Gn;
        end
    end
end
if nargin == 3
    if isempty(plt)
        plt = false;
    end
end


% Determine the number of degrees of freedom of the object when considering
% only the bilateral constraints.
ndof = nnu - ndof_lost;
if ndof == 0
    % Form closure by bilateral constraints
    % fprintf('The object is form-closed by the bilateral constraints. \n');
    % fprintf('No unilateral constraints are needed. \n');
    metric = ncU + 2 * ncB;
    flag = 1;
    return;
end

% Build n-d convex hull.  First, add the origin as the first column of G.
% If it is not in the convex hull, then the grasp is form closed.
G = [zeros(ndof,1)   G];
if ndof == 1
    % The one-d case is easy.
    minG = min(G);
    maxG = max(G);
    metric = min(abs(minG), abs(maxG));
    if metric > eps
        flag = 1;
    end
else
    % All higher-dimensional cases.
    K = convhulln(G');
    if min(min(K)) > 1
        flag = 1; % form closure exists
        
        % Loop through the boundaries and find the minimum distance, which
        % is the metric value.
        minDis = -1;
        for i = 1:size(K, 1)
            v = zeros(ndof, ndof);
            for j = 1:ndof
                v(:, j) = G(:, K(i, j));
            end
            [jctDis, jctNrml] = jct_dist(v);
            if jctDis < minDis || minDis == -1
                minDis = jctDis;
                minNrml = jctNrml;
            end
        end
        metric = minDis;
        % minDis has been found, and that's the value of the metric.
    end    
end

%Plot the data and convex hull when possible
if plt
    if ndof ~= 2  &&  ndof ~= 3
            warning('Plotting is supported when the object has two or three degrees of freedom.');
    else
        figure
        hold on
        grid on
        axis equal;
        if ndof == 2
            patch(G(1, K(:,1)), G(2, K(:,1)),'cyan');
            plot(G(1, :), G(2, :), 'r.', 'MarkerSize', 16);
            plot(0, 0, 'k.', 'MarkerSize', 20);
            xlabel('Fx');
            ylabel('Fy');
        elseif ndof == 3
            % Plot the convex hull of the columns of G
            trisurf(K, G(1, :), G(2, :), G(3, :), 'FaceColor', 'cyan', 'FaceAlpha', 0.5);
            plot3(G(1, :), G(2, :), G(3, :), 'r.', 'MarkerSize', 16);
            plot3(0, 0, 0, 'k.', 'MarkerSize', 20);
            xlabel('Fx');
            ylabel('Fy');
            zlabel('Mz');
        end
    end
    if metric > 0
        if ndof == 2
            % Plot the maximum inscribed circle
            radius = metric;
            theta = 0 : 0.01 : 2*pi;
            theta = [theta  0];
            patch(radius*cos(theta), radius*sin(theta), 'yellow');
            plot(minDis*cos(theta), minDis*sin(theta), 'm', 'LineWidth',2);
            plot(0, 0, 'k.', 'MarkerSize', 20);
            plot([0 1.5*metric*minNrml(1)], [0 1.5*metric*minNrml(2)], 'b', 'LineWidth',2);
        elseif ndof == 3
            % Plot the maximum inscribed sphere
            [x,y,z] = sphere;
            surf(metric*x, metric*y, metric*z, 'FaceColor','yellow', 'FaceAlpha',0.5, 'EdgeColor','none')  % sphere centered at origin
            plot3([0 2*metric*minNrml(1)], [0 2*metric*minNrml(2)], [0 2*metric*minNrml(3)], 'b', 'LineWidth', 2);
        end
    end
end
end

% This function computes the distance between the origin of the wrench
% space and a facet of the convex hull.  It works for any dimension.
function [dist, normal] = jct_dist(v)
   numPts = size(v,1);
   coeffs = v' \ ones(numPts,1);  % Solve for normal to hyper-plane through numPts points
   scale = norm(coeffs);
   normal = coeffs / scale;
   dist = 1.0/scale;
end
