function [frcCl, metricFrFrmCl] = forceClosure(G, J, type, mu)
% This code does a binary test for form closure and also returns a metric
% quantifying how strongly it satisfies frictional form closure.
%
%  Usage:  [frcCl, metricFrFrmCl] = forceClosure(G, J, type, mu)
%
% Input paramters:
%    G     - the Grasp matrix for the hand with the specified contacts
%    J     - the Jacobian matrix for the hand with the specified contacts
%    type  - a vector of contact types:
%               1 -->  point without friction
%               2 -->  hard finger contact
%               3 -->  soft finger contact
%    mu    - a vector of friction coefficients
%
% Output paramters:
%    frcCl - flag indicating whether or not the grasp has force closure:
%               -1 -->  not enough constraints for frict. form closure
%                0 -->  enough constraints, but still no force closure
%                1 -->  force closure exists
%    metricFrFrmCl - a measure of how far the grasp is from losing
%                         frictional form closure
%   positive value --> frictional form closure exists
%                0 -->  frictional form closure does not exist
%             -100 -->  frictional form closure could not be evaluated

% Written by Jeff Trinkle and Giuseppe Muscio, March 2013.

% Set default return values
frcCl = -2;
metricFrFrmCl = -100;

%==============================================
% Check input validity
n_c = length(type);
if length(mu) ~= n_c
    error('The length of "mu" and "type" must be the same.');
end

[n_nu, n_lambda] = size(G);
[n_q, n_lambdaJ] = size(J');
if n_lambdaJ ~= n_lambda
    error('The number of rows of J must equal the number of columns of G.');
end

% Get numbers of various contact types to check sizes of G and J
n_pwof = length(find(type == 1));   % number of frictionless contacts
n_hf = length(find(type == 2));     % number of hard-finger contacts
n_sf = length(find(type == 3));     % number of soft-finger contacts

switch n_nu
    % Determine n_nu, the number of dof of the unconstrained object
    case 3    % Planar case
        n_lambda_check = n_pwof + 2*n_hf;
        if n_sf > 0
            warning('Converting SF contacts to HF in planar problem.')
            n_lambda_check = n_lambda_check + 2*n_sf;
        end
        if n_lambda_check ~= n_lambda
            error('Number of columns of G and rows of J incompatible with contact "type" array.');
        end
    case 6    % Spatial case
        n_lambda_check = n_pwof + 3*n_hf + 4*n_sf;
        if n_lambda_check ~= n_lambda
            error('Number of columns of G and rows of J incompatible with contact "type" array.');
        end
    otherwise
        error('The number of rows of G must be 3 (planar grasps) or 6 (spatial grasps).');
end

% End of input checking
%=================================


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Since we got this far, frictional form closure is possible.  Start
% testing.
if n_lambda < n_nu
    metricFrFrmCl = 0;   % Not enough contact constraint for frictional form closure
    frcCl = -1;          % Don't even test force closure, since it's not possible
    warning('Not enough contact constraints for frictional form closure; ell < n_nu.');
    return;
end
if rank(G) ~= n_nu
    metricFrFrmCl = 0;   % Not enough contact constraint for frictional form closure
    frcCl = -1;    % Don't even test force closure, since it's not possible
    return;
end

% Construct some things necessar for testing frictional form closure, the
% F, e, and E matrices.
switch n_nu
    % Determine n_nu, the number of dof of the unconstrained object
    case 3    % Planar case
        [F, e, E] = FeE2d(n_c, type, mu);
    case 6    % Spatial case
        [F, e, E] = FeE3d(n_c, type, mu);
end
[Frows,Fcols] = size(F);


% Set up LP2.  Recall that LP2 tries to maximize d.  Matlab's linprog
% minimizes its objective value.  Thus we minimize -d to achieve the same
% affect.  Matlab also assumes all inqualities are formed as less than or
% equal.
f = [zeros(n_lambda, 1);
        -1      ];  % This gives us: minimize d
    
A2 = [ [-F             ones(Frows, 1)];  
       [ e                   0       ];  
       [zeros(1, Fcols)     -1       ]]; 

b2 = [zeros(Frows, 1)
         n_c
          0      ];
      
Aeq2 = [G   zeros(n_nu, 1)];
beq2 = [zeros(n_nu, 1)];

options = optimset('Display', 'off');

% Solve LP2.  The objective function value FVAL will be the
% optimal (-d).  Therefore dstar will be -FVAL 
[X2, FVAL2, EXITFLAG2] = linprog(f, A2, b2, Aeq2, beq2, [], [], [], options);
metricFrFrmCl = -FVAL2 * n_c;
% fprintf('Frictional form closure measure = %0.5g \n', -FVAL2);
if FVAL2 < -1e-6
%     fprintf('Frictional form closure exists.\n');
else
%     fprintf('Frictional form closure DOES NOT exist.\n');
    frcCl = -1;
    return;
end

%==========================================
% If we got this far, then the grasp has frictional form closure, so now
% finally test for force closure by solving LP3.
A3 =[[ -E   ones(n_c, 1) ];
     [  e       0      ]];

b3 = [zeros(n_c, 1);
          n_c];

Aeq3 = [G   zeros(n_nu, 1);
        J'  zeros(n_q, 1) ];
beq3 = [zeros(n_nu + n_q, 1)];

% Solve LP3
[X3, FVAL3, EXITFLAG3] = linprog(f,A3,b3,Aeq3,beq3,[],[],[],options);
if FVAL3 > 1e-6   % This is an arbitrary threshold.
%     fprintf('Force closure does not exist. \n');
    frcCl = 0;
else
%     fprintf('Force closure Exists. \n');
    frcCl = 1;
end
end


% Compute the matrices needed to formulate LP2 for planar grasps.
function [F, e, E] = FeE2d(n_c, type, mu)
F = [];
e = [];
E = [];
for i = 1 : n_c
    switch type(i)
        case 1
            Fi = [1];
            ei = [1];
        case 2
            alpha = atan2(mu(i),1);
            s_alpha = sin(alpha);
            c_alpha = cos(alpha);
            Fi = [ s_alpha  -c_alpha;
                s_alpha   c_alpha]; % For hard finger
            ei = [1 0];
    end
    F = blkdiag(F,  Fi);
    e = [e   ei];
    E = blkdiag(E,  ei);
end
end

% Compute the matrices needed to formulate LP2 for spatial grasps.
function [F, e, E] = FeE3d(n_c, type, mu)
F = [];
e = [];
E = [];
n_g = 11;  % Default number of facets of the linearized friction cone.
for i = 1 : n_c
    switch type(i)
        case 1
            Fi = [1];
            ei = [1];
        case 2
            alpha1 = atan2(mu(i),1);
            beta = 2 * pi / n_g;
            beta = beta * [1 : n_g]';
            Fi = [ones(n_g,1) * sin(alpha1)  cos(beta) * cos(alpha1)  sin(beta) * cos(alpha1)]; % For hard finger
            ei = [1 0 0];
        case 3
            ups = 1;   % Default value for torsional friction coefficient
            alpha1 = atan2(mu(i),1);
            beta = 2 * pi / n_g;
            beta = beta * [1 : n_g]';
            eta1 = [ones(n_g,1) * sin(alpha1)  cos(beta) * cos(alpha1)  sin(beta) * cos(alpha1)]; % For soft finger too
            Fi = [ eta1        zeros(n_g,1);
                [ups * mu(i)   0   0  -1] / norm([ups * mu(i)   0   0  -1]);
                [ups * mu(i)   0   0   1] / norm([ups * mu(i)   0   0  -1])];
            ei = [1 0 0 0];
    end
    F = blkdiag(F,  Fi);
    e = [e   ei];
    E = blkdiag(E,  ei);
end
end