function [flag, metric] = formClosure(Gn, B)
% Form closure test from Grasping chapter of Springer Robotics Handbook,
% edition 2. 
%
% This test handles two cases: (1) the object is constrained only by
% unilateral contacts (captured by Gn) and (2) the object is constrained 
% also by bilateral constraints (captured by B).  The analysis is done on
% the matrix G, which is computed from Gn and B and plays the role of Gn 
% after the bilateral constraints are accounted for.
%
%   usage:  [flag, metric] = formClosure(Gn, B)
%
% Input parameters:
%     Gn     -- matrix whose columns are the unit normal wrenches. 
%                 Unilateral constraints:   Gn'*nu >= 0
%     B      -- matrix whose columns are the unit wrenches for bilateral
%                 constraints.  Bilateral constraints:     B'*nu = 0
% Output parameters:
%     metric -- measure of how far grasp is from losing form closure
%                    metric = dstar * number_unilateral_contacts
%                 where number_of_unilateral contacts is the number of
%                 columns in Gn plus 2*(number of columns in B).
%     flag = -2  -->  grasp doesn't have enough contacts for form closure
%          = -1  -->  G not full row rank, so form closure doesn't exist
%          =  0  -->  grasp does not have form closure for other reasons
%          =  1  -->  grasp has form closure

% Written by Jeff Trinkle and Giuseppe Muscio, February 2013.


% Get the number of rows and columns of Gn
% nnu --> number of degrees of freedom of object
% ncU --> number of unilateral contacts
[nnu, ncU] = size(Gn);
if nnu == 0
    error('You passed an empty wrench matrix, Gn, into formClosure()');
end

switch nargin
    case 1
        ncB = 0;           % Number of columns of B
        ndof_lost = 0;     % Since B is empty, no lost dofs
        G = Gn;            % Gn is not changed
    case 2
        if isempty(B)
            ncB = 0;       % Number of columns of B
            ndof_lost = 0; % Since B is empty, no lost dofs
            G = Gn;        % Gn is not changed
        else
            [~, ncB] = size(B);   % Number of columns of B
            ndof_lost = rank(B);  % Lose dofs from constraint:  B' * nu = 0
            % A_B is defined as the transpose of the matrix whose columns
            % span the null space of B' and thus A_B*B = 0.
            A_B = null(B')';
            % G is the grasp matrix \overline{G} defined in the Handbook
            G = A_B*Gn;
        end
    otherwise
end
ndof = nnu - ndof_lost;


if ndof == 0
    % Form closure by bilateral constraints
    % fprintf('The object is form-closed by the bilateral constraints. \n');
    % fprintf('No unilateral constraints are needed. \n');
    metric = ncU + 2 * ncB;
    flag = 1;
    return;
    
else
    % Bilateral constraints do not provide form closure, i.e., ndof > 0
    if ncU <= ndof;
        % Not enough unilateral contacts for form closure
        % fprintf('Form closure DOES NOT exist, because there are NOT MORE \n');
        % fprintf('unilateral contacts than object degrees of freedom. \n');
        % fprintf('     number of degress of freedom = %2d. \n', ndof);
        % fprintf('     number of unilateral contacts = %2d. \n', nc);
        metric = 0;
        flag = -2;
        return;
    end
    
    if rank(G) < ndof
        % form closure is not possible
        % fprintf('Form closure not possible. \n');
        % fprintf('Gn is not full row rank.  Rank(Gn)= %2d, but it has %2d rows.', nr);
        metric = 0;
        flag = -1;
        return;
    else
        % Set up LP1 from page 684 of Springer Robotics handbook, edition 1.
        % All equation numbers below are for edition 1.
        % Since the Matlab LP solver assumes minimization, negate objective
        % function to effect maximization.
        f = [zeros(ncU,1);  % Equation (28.36)
            -1;];     % This gives minimize -d <==> maximize d
        
        A = [ [-eye(ncU,ncU)      ones(ncU,1) ];    % Equation (28.38), left
              [ ones(1,ncU)            0      ];    % Equation (28.40), left
              [ zeros(1,ncU)          -1      ] ];  % Equation (28.39), left
        
        b = [zeros(ncU,1)     % Equation (28.38), right
               ncU            % Equation (28.40), right
                0     ];      % Equation (28.39), right
        
        Aeq = [G   zeros(ndof,1)];     % Equation (28.37), left
        beq =   zeros(ndof,1) ;        % Equation (28.37), right
        
        
        % Call Matlab's linear program solver
        options = optimset('Display','off');
        [~,FVAL] = linprog(f,A,b,Aeq,beq,[],[],[],options);
        if FVAL < -1e10 * eps   % linprog default accuracy is 1e-6, about 1e10 *eps
            % fprintf('Form closure exists!!!  Closure metric = %0.5g \n', -FVAL);
            num_conts = ncU + 2 * ncB;
            metric = -FVAL * num_conts;
            flag = 1;
        else
            % fprintf('Form closure DOES NOT exist.\n');
            metric = 0;
            flag = 0;
        end
        return;
    end
end
